/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%                        PPPP    AAA   L      M   M                           %
%                        P   P  A   A  L      MM MM                           %
%                        PPPP   AAAAA  L      M M M                           %
%                        P      A   A  L      M   M                           %
%                        P      A   A  LLLLL  M   M                           %
%                                                                             %
%                                                                             %
%                          Read/Write Palm Pixmap.                            %
%                                                                             %
%                                                                             %
%                              Software Design                                %
%                            Christopher R. Hawks                             %
%                               December 2001                                 %
%                                                                             %
%                                                                             %
%  Copyright (C 2001 ImageMagick Studio LLC, a non-profit organization dedicated  %
%  to making software imaging solutions freely available.                     %
%                                                                             %
%  Permission is hereby granted, free of charge, to any person obtaining a    %
%  copy of this software and associated documentation files ("ImageMagick",   %
%  to deal in ImageMagick without restriction, including without limitation   %
%  the rights to use, copy, modify, merge, publish, distribute, sublicense,   %
%  and/or sell copies of ImageMagick, and to permit persons to whom the       %
%  ImageMagick is furnished to do so, subject to the following conditions:    %
%                                                                             %
%  The above copyright notice and this permission notice shall be included in %
%  all copies or substantial portions of ImageMagick.                         %
%                                                                             %
%  The software is provided "as is", without warranty of any kind, express or %
%  implied, including but not limited to the warranties of merchantability,   %
%  fitness for a particular purpose and noninfringement.  In no event shall   %
%  ImageMagick Studio be liable for any claim, damages or other liability,    %
%  whether in an action of contract, tort or otherwise, arising from, out of  %
%  or in connection with ImageMagick or the use or other dealings in          %
%  ImageMagick.                                                               %
%                                                                             %
%  Except as contained in this notice, the name of the ImageMagick Studio     %
%  shall not be used in advertising or otherwise to promote the sale, use or  %
%  other dealings in ImageMagick without prior written authorization from the %
%  ImageMagick Studio.                                                        %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Based on pnmtopalm by Bill Janssen and ppmtobmp by Ian Goldberg.
%
*/

/*
  Include declarations.
*/
#include "magick/studio.h"
#include "magick/attribute.h"
#include "magick/blob.h"
#include "magick/color.h"
#include "magick/color_private.h"
#include "magick/colorspace.h"
#include "magick/constitute.h"
#include "magick/error.h"
#include "magick/image.h"
#include "magick/list.h"
#include "magick/magick.h"
#include "magick/memory_.h"
#include "magick/paint.h"
#include "magick/quantize.h"
#include "magick/static.h"
#include "magick/string_.h"
#include "magick/utility.h"

/*
  Define declarations.
*/
#define PALM_IS_COMPRESSED_FLAG  0x8000
#define PALM_HAS_COLORMAP_FLAG  0x4000
#define PALM_HAS_TRANSPARENCY_FLAG  0x2000
#define PALM_IS_INDIRECT  0x1000
#define PALM_IS_FOR_SCREEN  0x0800
#define PALM_IS_DIRECT_COLOR  0x0400

#define PALM_COMPRESSION_SCANLINE  0x00
#define PALM_COMPRESSION_RLE  0x01
#define PALM_COMPRESSION_NONE  0xFF

/*
 The 256 color system palette for Palm Computing Devices.
*/
static unsigned char
  PalmPalette[256][3] =
  {
    {255, 255,255}, {255, 204,255}, {255, 153,255}, {255, 102,255},
    {255,  51,255}, {255,   0,255}, {255, 255,204}, {255, 204,204},
    {255, 153,204}, {255, 102,204}, {255,  51,204}, {255,   0,204},
    {255, 255,153}, {255, 204,153}, {255, 153,153}, {255, 102,153},
    {255,  51,153}, {255,   0,153}, {204, 255,255}, {204, 204,255},
    {204, 153,255}, {204, 102,255}, {204,  51,255}, {204,   0,255},
    {204, 255,204}, {204, 204,204}, {204, 153,204}, {204, 102,204},
    {204,  51,204}, {204,   0,204}, {204, 255,153}, {204, 204,153},
    {204, 153,153}, {204, 102,153}, {204,  51,153}, {204,   0,153},
    {153, 255,255}, {153, 204,255}, {153, 153,255}, {153, 102,255},
    {153,  51,255}, {153,   0,255}, {153, 255,204}, {153, 204,204},
    {153, 153,204}, {153, 102,204}, {153,  51,204}, {153,   0,204},
    {153, 255,153}, {153, 204,153}, {153, 153,153}, {153, 102,153},
    {153,  51,153}, {153,   0,153}, {102, 255,255}, {102, 204,255},
    {102, 153,255}, {102, 102,255}, {102,  51,255}, {102,   0,255},
    {102, 255,204}, {102, 204,204}, {102, 153,204}, {102, 102,204},
    {102,  51,204}, {102,   0,204}, {102, 255,153}, {102, 204,153},
    {102, 153,153}, {102, 102,153}, {102,  51,153}, {102,   0,153},
    { 51, 255,255}, { 51, 204,255}, { 51, 153,255}, { 51, 102,255},
    { 51,  51,255}, { 51,   0,255}, { 51, 255,204}, { 51, 204,204},
    { 51, 153,204}, { 51, 102,204}, { 51,  51,204}, { 51,   0,204},
    { 51, 255,153}, { 51, 204,153}, { 51, 153,153}, { 51, 102,153},
    { 51,  51,153}, { 51,   0,153}, {  0, 255,255}, {  0, 204,255},
    {  0, 153,255}, {  0, 102,255}, {  0,  51,255}, {  0,   0,255},
    {  0, 255,204}, {  0, 204,204}, {  0, 153,204}, {  0, 102,204},
    {  0,  51,204}, {  0,   0,204}, {  0, 255,153}, {  0, 204,153},
    {  0, 153,153}, {  0, 102,153}, {  0,  51,153}, {  0,   0,153},
    {255, 255,102}, {255, 204,102}, {255, 153,102}, {255, 102,102},
    {255,  51,102}, {255,   0,102}, {255, 255, 51}, {255, 204, 51},
    {255, 153, 51}, {255, 102, 51}, {255,  51, 51}, {255,   0, 51},
    {255, 255,  0}, {255, 204,  0}, {255, 153,  0}, {255, 102,  0},
    {255,  51,  0}, {255,   0,  0}, {204, 255,102}, {204, 204,102},
    {204, 153,102}, {204, 102,102}, {204,  51,102}, {204,   0,102},
    {204, 255, 51}, {204, 204, 51}, {204, 153, 51}, {204, 102, 51},
    {204,  51, 51}, {204,   0, 51}, {204, 255,  0}, {204, 204,  0},
    {204, 153,  0}, {204, 102,  0}, {204,  51,  0}, {204,   0,  0},
    {153, 255,102}, {153, 204,102}, {153, 153,102}, {153, 102,102},
    {153,  51,102}, {153,   0,102}, {153, 255, 51}, {153, 204, 51},
    {153, 153, 51}, {153, 102, 51}, {153,  51, 51}, {153,   0, 51},
    {153, 255,  0}, {153, 204,  0}, {153, 153,  0}, {153, 102,  0},
    {153,  51,  0}, {153,   0,  0}, {102, 255,102}, {102, 204,102},
    {102, 153,102}, {102, 102,102}, {102,  51,102}, {102,   0,102},
    {102, 255, 51}, {102, 204, 51}, {102, 153, 51}, {102, 102, 51},
    {102,  51, 51}, {102,   0, 51}, {102, 255,  0}, {102, 204,  0},
    {102, 153,  0}, {102, 102,  0}, {102,  51,  0}, {102,   0,  0},
    { 51, 255,102}, { 51, 204,102}, { 51, 153,102}, { 51, 102,102},
    { 51,  51,102}, { 51,   0,102}, { 51, 255, 51}, { 51, 204, 51},
    { 51, 153, 51}, { 51, 102, 51}, { 51,  51, 51}, { 51,   0, 51},
    { 51, 255,  0}, { 51, 204,  0}, { 51, 153,  0}, { 51, 102,  0},
    { 51,  51,  0}, { 51,   0,  0}, {  0, 255,102}, {  0, 204,102},
    {  0, 153,102}, {  0, 102,102}, {  0,  51,102}, {  0,   0,102},
    {  0, 255, 51}, {  0, 204, 51}, {  0, 153, 51}, {  0, 102, 51},
    {  0,  51, 51}, {  0,   0, 51}, {  0, 255,  0}, {  0, 204,  0},
    {  0, 153,  0}, {  0, 102,  0}, {  0,  51,  0}, { 17,  17, 17},
    { 34,  34, 34}, { 68,  68, 68}, { 85,  85, 85}, {119, 119,119},
    {136, 136,136}, {170, 170,170}, {187, 187,187}, {221, 221,221},
    {238, 238,238}, {192, 192,192}, {128,   0,  0}, {128,   0,128},
    {  0, 128,  0}, {  0, 128,128}, {  0,   0,  0}, {  0,   0,  0},
    {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0},
    {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0},
    {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0},
    {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0},
    {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0},
    {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0}, {  0,   0,  0}
  };

/*
  Forward declarations.
*/
static MagickBooleanType
  WritePALMImage(const ImageInfo *,Image *);

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   F i n d C o l o r                                                         %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  FindColor() returns the index of the matching entry from PalmPalette for a
%  given PixelPacket.
%
%  The format of the FindColor method is:
%
%      int FindColor(PixelPacket *pixel)
%
%  A description of each parameter follows:
%
%    o int: the index of the matching color or -1 if not found/
%
%    o pixel: a pointer to the PixelPacket to be matched.
%
%
*/
static int FindColor(PixelPacket *pixel)
{
  register long
    i;

  for (i=0; i < 256; i++)
    if (ScaleQuantumToChar(pixel->red) == PalmPalette[i][0] &&
        ScaleQuantumToChar(pixel->green) == PalmPalette[i][1] &&
        ScaleQuantumToChar(pixel->blue) == PalmPalette[i][2])
      return(i);
  return(-1);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   R e a d P A L M I m a g e                                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  ReadPALMImage() reads an image of raw bites in LSB order and returns it.  It
%  allocates the memory necessary for the new Image structure and returns a
%  pointer to the new image.
%
%  The format of the ReadPALMImage method is:
%
%      Image *ReadPALMImage(const ImageInfo *image_info,
%        ExceptionInfo *exception)
%
%  A description of each parameter follows:
%
%    o image_info: Specifies a pointer to an ImageInfo structure.
%
%    o exception: return any errors or warnings in this structure.
%
%
*/
static Image *ReadPALMImage(const ImageInfo *image_info,
  ExceptionInfo *exception)
{
  Image
    *image;

  IndexPacket
    index;

  long
    y;

  MagickBooleanType
    status;

  register IndexPacket
    *indexes;

  register long
    i,
    x;

  register PixelPacket
    transpix,
    *q;

  ssize_t
    count;

  unsigned char
    *lastrow,
    *one_row,
    *ptr;

  unsigned long
    bytes_per_row,
    flags,
    bits_per_pixel,
    version,
    nextDepthOffset,
    transparentIndex,
    compressionType,
    byte,
    mask,
    redbits,
    greenbits,
    bluebits,
    pad,
    size,
    bit;

  unsigned short
    color16;

  /*
    Open image file.
  */
  image=AllocateImage(image_info);
  status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
  if (status == MagickFalse)
    {
      DestroyImageList(image);
      return((Image *) NULL);
    }
  image->columns=ReadBlobMSBShort(image);
  image->rows=ReadBlobMSBShort(image);
  bytes_per_row=ReadBlobMSBShort(image);
  flags=ReadBlobMSBShort(image);
  bits_per_pixel=ReadBlobByte(image);
  version=ReadBlobByte(image);
  nextDepthOffset=ReadBlobMSBShort(image);
  transparentIndex=ReadBlobByte(image);
  compressionType=ReadBlobByte(image);
  pad=ReadBlobMSBShort(image);
  /*
    Initialize image colormap.
  */
  if ((bits_per_pixel < 16) &&
      !AllocateImageColormap(image,1L << bits_per_pixel))
    ThrowReaderException(ResourceLimitError,"MemoryAllocationFailed");
  if (bits_per_pixel == 16)  /* Direct Color */
    {
      redbits=ReadBlobByte(image);  /* # of bits of red */
      greenbits=ReadBlobByte(image);  /* # of bits of green */
      bluebits=ReadBlobByte(image);  /* # of bits of blue */
      ReadBlobByte(image);  /* reserved by Palm */
      ReadBlobByte(image);  /* reserved by Palm */
      transpix.red=MaxRGB*ReadBlobByte(image)/31;
      transpix.green=MaxRGB*ReadBlobByte(image)/63;
      transpix.blue=MaxRGB*ReadBlobByte(image)/31;
    }
  if (bits_per_pixel == 8)
    {
      if (flags & PALM_HAS_COLORMAP_FLAG)
        {
          count=ReadBlobMSBShort(image);
          for (i=0; i < (long) count; i++)
          {
            ReadBlobByte(image);
            image->colormap[255-i].red=ScaleCharToQuantum(ReadBlobByte(image));
            image->colormap[255-i].green=
              ScaleCharToQuantum(ReadBlobByte(image));
            image->colormap[255-i].blue=ScaleCharToQuantum(ReadBlobByte(image));
        }
      }
    else
      {
        for (i = 0; i < (long) (1L << bits_per_pixel); i++)
        {
          image->colormap[255-i].red=ScaleCharToQuantum(PalmPalette[i][0]);
          image->colormap[255-i].green=ScaleCharToQuantum(PalmPalette[i][1]);
          image->colormap[255-i].blue=ScaleCharToQuantum(PalmPalette[i][2]);
        }
      }
    }
  if (flags & PALM_IS_COMPRESSED_FLAG)
    size=ReadBlobMSBShort(image);
  image->storage_class=DirectClass;
  if (bits_per_pixel < 16)
    {
      image->storage_class=PseudoClass;
      image->depth=8;
    }
  one_row=(unsigned char *) AcquireMagickMemory(bytes_per_row);
  if (one_row == (unsigned char *) NULL)
    ThrowReaderException(ResourceLimitError,"MemoryAllocationFailed");
  lastrow=(unsigned char *) NULL;
  if (compressionType == PALM_COMPRESSION_SCANLINE)
    lastrow=(unsigned char *) AcquireMagickMemory(bytes_per_row);
  mask=(1l << bits_per_pixel)-1;
  for (y = 0; y < (long) image->rows; y++)
  {
    if ((flags & PALM_IS_COMPRESSED_FLAG) == 0)
      {
        image->compression=NoCompression;
        (void) ReadBlob(image,bytes_per_row,one_row);
      }
    else
      {
        if (compressionType == PALM_COMPRESSION_RLE)
          {
            image->compression=RLECompression;
            for (i=0; i < (long) bytes_per_row; )
            {
              count=ReadBlobByte(image);
              byte=ReadBlobByte(image);
              (void) ResetMagickMemory(one_row+i,(int) byte,count);
              i+=count;
            }
        }
      else
        if (compressionType == PALM_COMPRESSION_SCANLINE)
          {
            image->compression=FaxCompression;
            for (i=0; i < (long) bytes_per_row; i+=8)
            {
              count=ReadBlobByte(image);
              byte=Min(bytes_per_row-i,8);
              for (bit=0; bit < byte; bit++)
              {
                if ((y == 0) || (count & (1 << (7 - bit))))
                  one_row[i+bit]=ReadBlobByte(image);
                else
                  one_row[i+bit]=lastrow[i+bit];
              }
            }
            CopyMagickMemory(lastrow, one_row, bytes_per_row);
          }
      }
    ptr=one_row;
    q=SetImagePixels(image,0,y,image->columns,1);
    if (q == (PixelPacket *) NULL)
      break;
    indexes=GetIndexes(image);
    if (bits_per_pixel == 16)
      {
        for (x=0; x < (long) image->columns; x++)
        {
          color16=(*ptr++ << 8);
          color16|=(*ptr++);
          q->red=((MaxRGB*(color16 >> 11) & 0x1f))/31;
          q->green=((MaxRGB*(color16 >> 5) & 0x3f))/63;
          q->blue=(MaxRGB*(color16 & 0x1f))/31;
          q->opacity=OpaqueOpacity;
          q++;
        }
      }
    else
      {
        bit=8-bits_per_pixel;
        for (x=0; x < (long) image->columns; x++)
        {
          index=(IndexPacket) (mask-(((*ptr) & (mask << bit)) >> bit));
          indexes[x]=index;
          *q++=image->colormap[index];
          if (bit)
            bit-=bits_per_pixel;
          else
            {
              ptr++;
              bit=8-bits_per_pixel;
            }
        }
        if (SyncImagePixels(image) == MagickFalse)
          break;
      }
    }
  if (flags & PALM_HAS_TRANSPARENCY_FLAG)
    {
      if (bits_per_pixel == 16)
        TransparentImage(image,transpix,TransparentOpacity);
      else
        TransparentImage(image,image->colormap[mask-transparentIndex],
          TransparentOpacity);
    }
  one_row=(unsigned char *) RelinquishMagickMemory(one_row);
  if (compressionType == PALM_COMPRESSION_SCANLINE)
    lastrow=(unsigned char *) RelinquishMagickMemory(lastrow);
  CloseBlob(image);
  return(GetFirstImageInList(image));
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   R e g i s t e r P A L M I m a g e                                         %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  RegisterPALMImage() adds attributes for the PALM image format to the list of
%  supported formats.  The attributes include the image format tag, a method to
%  read and/or write the format, whether the format supports the saving of more
%  than one frame to the same file or blob, whether the format supports native
%  in-memory I/O, and a brief description of the format.
%
%  The format of the RegisterPALMImage method is:
%
%      RegisterPALMImage(void)
%
*/
ModuleExport void RegisterPALMImage(void)
{
  MagickInfo
    *entry;

  entry=SetMagickInfo("PALM");
  entry->decoder=(DecoderHandler *) ReadPALMImage;
  entry->encoder=(EncoderHandler *) WritePALMImage;
  entry->adjoin=MagickFalse;
  entry->seekable_stream=MagickTrue;
  entry->description=AcquireString("Palm pixmap");
  entry->module=AcquireString("PALM");
  (void) RegisterMagickInfo(entry);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   U n r e g i s t e r P A L M I m a g e                                     %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Method UnregisterPALMImage removes format registrations made by the
%  PALM module from the list of supported formats.
%
%  The format of the UnregisterPALMImage method is:
%
%      UnregisterPALMImage(void)
%
*/
ModuleExport void UnregisterPALMImage(void)
{
  UnregisterMagickInfo("PALM");
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   W r i t e P A L M I m a g e                                               %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  WritePALMImage() writes an image of raw bits in LSB order to a file.
%
%  The format of the WritePALMImage method is:
%
%      MagickBooleanType WritePALMImage(const ImageInfo *image_info,Image *image)
%
%  A description of each parameter follows.
%
%    o image_info: Specifies a pointer to an ImageInfo structure.
%
%    o image:  A pointer to a Image structure.
%
%
*/
static MagickBooleanType WritePALMImage(const ImageInfo *image_info,Image *image)
{
  int
    y;

  const ImageAttribute
    *attribute;

  Image
    *map;

  ExceptionInfo
    exception;

  MagickBooleanType
    status;

  QuantizeInfo
    quantize_info;

  register IndexPacket
    *indexes;

  register long
    x;

  register PixelPacket
    transpix,
    *p;

  unsigned char
    bit,
    byte,
    color,
    *lastrow,
    *one_row,
    *ptr,
    version = 0;

  unsigned int
    transparentIndex = 0;

  unsigned long
    count,
    bits_per_pixel,
    bytes_per_row;

  unsigned short
    offset,
    color16,
    flags = 0;


  /*
    Open output image file.
  */
  status=OpenBlob(image_info,image,WriteBinaryBlobMode,&exception);
  if (status == MagickFalse)
    return(status);
  GetExceptionInfo(&exception);
  attribute=GetImageAttribute(image, "Comment");
  if (attribute != (ImageAttribute *)NULL)
    if (LocaleCompare("COLORMAP",attribute->value) == 0)
      flags|=PALM_HAS_COLORMAP_FLAG;
  count = GetNumberColors(image, NULL, &exception);
  for (bits_per_pixel=1;  (1UL << bits_per_pixel) < count;  bits_per_pixel*=2);
  if (image_info->depth > 100)
    bits_per_pixel=image_info->depth-100;
  if (bits_per_pixel < 16)
    SetImageColorspace(image,image->colorspace);
  if (bits_per_pixel < 8)
    {
      SetImageColorspace(image,GRAYColorspace);
      (void) SortColormapByIntensity(image);
    }
  if (bits_per_pixel > 8)
    flags|=PALM_IS_DIRECT_COLOR;
  (void) WriteBlobMSBShort(image,image->columns);  /* width */
  (void) WriteBlobMSBShort(image,image->rows );  /* height */
  bytes_per_row=((image->columns+(16/bits_per_pixel-1))/(16/bits_per_pixel))*2;
  (void) WriteBlobMSBShort(image, bytes_per_row);
  if ((image->compression == RLECompression) ||
      (image->compression == FaxCompression))
    flags|=PALM_IS_COMPRESSED_FLAG;
  (void) WriteBlobMSBShort(image, flags);
  (void) WriteBlobByte(image, bits_per_pixel);
  if (bits_per_pixel > 1)
    version=1;
  if ((image->compression == RLECompression) ||
      (image->compression == FaxCompression))
    version=2;
  (void) WriteBlobByte(image,version);
  (void) WriteBlobMSBShort(image,0);  /* offset */
  (void) WriteBlobByte(image,transparentIndex);
  if (image->compression == RLECompression)
    (void) WriteBlobByte(image,PALM_COMPRESSION_RLE);
  else
    if (image->compression == FaxCompression)
      (void) WriteBlobByte(image,PALM_COMPRESSION_SCANLINE);
    else
      (void) WriteBlobByte(image,PALM_COMPRESSION_NONE);
  (void) WriteBlobMSBShort(image,0);  /* reserved */
  offset=16;
  if (bits_per_pixel == 16)
    {
      WriteBlobByte(image,5);  /* # of bits of red */
      WriteBlobByte(image,6);  /* # of bits of green */
      WriteBlobByte(image,5);  /* # of bits of blue */
      WriteBlobByte(image,0);  /* reserved by Palm */
      WriteBlobMSBLong(image,0);  /* no transparent color, YET */
      offset+=8;
    }
  if (bits_per_pixel == 8)
    {
      if (flags & PALM_HAS_COLORMAP_FLAG)  /* Write out colormap */
        {
          GetQuantizeInfo(&quantize_info);
          quantize_info.dither=image_info->dither;
          quantize_info.number_colors=image->colors;
          (void) QuantizeImage(&quantize_info,image);
          WriteBlobMSBShort(image,image->colors);
          for (count = 0; count < image->colors; count++)
          {
            WriteBlobByte(image,count);
            WriteBlobByte(image,ScaleQuantumToChar(image->colormap[count].red));
            WriteBlobByte(image,
              ScaleQuantumToChar(image->colormap[count].green));
            WriteBlobByte(image,
              ScaleQuantumToChar(image->colormap[count].blue));
          }
          offset+=2+count*4;
        }
    else  /* Map colors to Palm standard colormap */
      {
        map=ConstituteImage(256,1,"RGB",CharPixel,&PalmPalette,&exception);
        SetImageColorspace(map,map->colorspace);
        MapImage(image,map,image_info->dither);
        for (y=0; y < (long) image->rows; y++)
        {
          p=GetImagePixels(image,0,y,image->columns,1);
          indexes=GetIndexes(image);
          for (x=0; x < (long) image->columns; x++)
            indexes[x]=FindColor(&image->colormap[indexes[x]]);
        }
        DestroyImage(map);
      }
    }
  if (flags & PALM_IS_COMPRESSED_FLAG)
    WriteBlobMSBShort(image,0);  /* fill in size later */
  lastrow=(unsigned char *) NULL;
  if (image->compression == FaxCompression)
    lastrow=(unsigned char *) AcquireMagickMemory(bytes_per_row);
  one_row=(unsigned char *) AcquireMagickMemory(bytes_per_row);
  if (one_row == (unsigned char *) NULL)
    ThrowWriterException(ResourceLimitError,"MemoryAllocationFailed");
  for (y=0; y < (int) image->rows; y++)
  {
    ptr=one_row;
    (void) ResetMagickMemory(ptr,0,bytes_per_row);
    p=GetImagePixels(image,0,y,image->columns,1);
    if (p == (PixelPacket *) NULL)
      break;
    indexes=GetIndexes(image);
    if (bits_per_pixel == 16)
      {
        for (x=0; x < (int) image->columns; x++)
        {
          color16=(((31*p->red)/MaxRGB) << 11) | (((63*p->green)/MaxRGB) << 5) |
            ((31*p->blue)/MaxRGB);
          if (p->opacity == TransparentOpacity)
            {
              transpix.red=p->red;
              transpix.green=p->green;
              transpix.blue=p->blue;
              transpix.opacity=p->opacity;
              flags|=PALM_HAS_TRANSPARENCY_FLAG;
            }
          *ptr++=(color16 >> 8) & 0xff;
          *ptr++=color16 & 0xff;
          p++;
        }
      }
    else
      {
        byte=0x00;
        bit=(unsigned char) (8-bits_per_pixel);
        for (x=0; x < (int) image->columns; x++)
        {
          if (bits_per_pixel >= 8)
            color=(unsigned char) indexes[x];
          else
            color=(unsigned char) (indexes[x]*((1 << bits_per_pixel)-1)/
              Max(image->colors-1,1));
          byte|=color << bit;
          if (bit != 0)
            bit-=bits_per_pixel;
          else
            {
              *ptr++=byte;
              byte=0x00;
              bit=(unsigned char) (8-bits_per_pixel);
            }
        }
        if ((image->columns % (8/bits_per_pixel)) != 0)
          *ptr++=byte;
      }
    if (image->compression == RLECompression)
      {
        x=0;
        while (x < (long) bytes_per_row)
        {
          byte=one_row[x];
          count=1;
          while ((one_row[++x] == byte) && (count < 255) &&
                 (x < (long) bytes_per_row))
            count++;
          (void) WriteBlobByte(image,count);
          (void) WriteBlobByte(image,byte);
        }
      }
    else
      if (image->compression == FaxCompression)
        {
          char
            tmpbuf[8],
            *tptr;

          for (x = 0;  x < (long) bytes_per_row;  x += 8)
          {
            tptr = tmpbuf;
            for (bit = 0, byte = 0; bit < Min(8, bytes_per_row - x); bit++)
            {
              if ((y == 0) || (lastrow[x + bit] != one_row[x + bit]))
                {
                byte |= (1 << (7 - bit));
                *tptr++ = one_row[x + bit];
              }
            }
            (void) WriteBlobByte(image, byte);
            (void) WriteBlob(image,tptr-tmpbuf,(unsigned char *) tmpbuf);
          }
          (void) CopyMagickMemory(lastrow,one_row,bytes_per_row);
        }
      else
        (void) WriteBlob(image,bytes_per_row,one_row);
    }
  if (flags & PALM_HAS_TRANSPARENCY_FLAG)
    {
      SeekBlob(image,6,SEEK_SET);
      (void) WriteBlobMSBShort(image,flags);
      SeekBlob(image,12,SEEK_SET);
      (void) WriteBlobByte(image,transparentIndex);  /* trans index */
    }
  if (bits_per_pixel == 16)
    {
      SeekBlob(image,20,SEEK_SET);
      WriteBlobByte(image,0);  /* reserved by Palm */
      WriteBlobByte(image,(31*transpix.red)/MaxRGB);
      WriteBlobByte(image,(63*transpix.green)/MaxRGB);
      WriteBlobByte(image,(31*transpix.blue)/MaxRGB);
    }
  if (flags & PALM_IS_COMPRESSED_FLAG)  /* fill in size now */
    {
      SeekBlob(image,offset,SEEK_SET);
      (void) WriteBlobMSBShort(image,GetBlobSize(image)-16);
    }
  CloseBlob(image);
  one_row=(unsigned char *) RelinquishMagickMemory(one_row);
  if (image->compression == FaxCompression)
    lastrow=(unsigned char *) RelinquishMagickMemory(lastrow);
  DestroyExceptionInfo(&exception);
  return(MagickTrue);
}
